#include <HighGUI.h>

extern int UnDistortOne(float xr, float yr, float *xu, float *yu, float *DistortionMatrix);
extern void DistortOne(float xu, float yu, float *xr, float *yr, float *DistortionMatrix);

//
//	Distort an array of points to obtain the loactions on the real distorted image
//
void DistortPoints(CvPoint2D32f *undist, CvPoint2D32f *dist, int npoints, float *CameraMatrix, float *DistortionMatrix)
{
	int i;
	float xu, yu;	// undistorted
	float xr, yr;	// real
	
	for(i=0; i < npoints; ++i)
	{
		xu = (undist[i].x-CameraMatrix[2])/CameraMatrix[0];
		yu = (undist[i].y-CameraMatrix[5])/CameraMatrix[4];
		
		DistortOne(xu, yu, &xr, &yr, DistortionMatrix);
		
		dist[i].x = xr*CameraMatrix[0] + CameraMatrix[2];
		dist[i].y = yr*CameraMatrix[4] + CameraMatrix[5];
	}
}

//
//	Undistort an array of points. Ie I mark points on the distorted (real) image and obtain
//	the real points for a perfect lens.
//
void UnDistortPoints(CvPoint2D32f *dist, CvPoint2D32f *undist, int npoints, float *CameraMatrix, float *DistortionMatrix)
{
	int i;
	float xu, yu;	// undistorted
	float xr, yr;	// real
	
	for(i=0; i < npoints; ++i)
	{
		xr = (dist[i].x-CameraMatrix[2])/CameraMatrix[0];
		yr = (dist[i].y-CameraMatrix[5])/CameraMatrix[4];
		
		UnDistortOne(xr, yr, &xu, &yu, DistortionMatrix);
		
		undist[i].x = xu*CameraMatrix[0] + CameraMatrix[2];
		undist[i].y = yu*CameraMatrix[4] + CameraMatrix[5];
	}
}

static void DistortOne(float xu, float yu, float *xr, float *yr, float *DistortionMatrix)
{
	float rr = xu*xu+yu*yu;
	float xy2 = 2.*xu*yu;
	
	float rd = 1. + rr*(DistortionMatrix[0]+DistortionMatrix[1]*rr);
	
	*xr = xu*rd + DistortionMatrix[2]*xy2 + DistortionMatrix[3]*(rr+2.*xu*xu);
	*yr = yu*rd + DistortionMatrix[3]*xy2 + DistortionMatrix[2]*(rr+2.*yu*yu);
}

//
//	Global parameters for the error evaluation function
//
static double xd, yd;
static float *dm;

extern "C"
{
#include "hooke.h"
}

//
//	Error evaluation function
//
static double delta(double *params)
{
	float xr, yr;
	
	DistortOne((float)params[0], (float)params[1], &xr, &yr, dm);
	
	return (xd - xr)*(xd - xr)+(yd - yr)*(yd - yr);
}

int UnDistortOne(float xr, float yr, float *xu, float *yu, float *DistortionMatrix)
{
	double	   startpt[2], endpt[2];
	int		   nvars, itermax, niter;
	double	   rho, epsilon;

	//
	//	Save the fixed parameters
	//
	xd = xr;
	yd = yr;
	dm = DistortionMatrix;

	//
	//	Set the solutor parameters
	//
	itermax = 5000;		// max # of iterations
	rho     = 0.9;		// stepsize geometric shrink
	epsilon = 1E-20;	// ending value of stepsize
	nvars	= 2;		// number of variables
	
	//
	//	Starting guess
	//
	startpt[0] = xr;
	startpt[1] = yr;
	
	//
	//	Solve
	//
	niter = hooke(nvars, startpt, endpt, rho, epsilon, itermax, delta);
	
	*xu = endpt[0];
	*yu = endpt[1];
	
	//
	//	Return number of iterations
	//
	return niter;
}